# SPDX-License-Identifier: GPL-3.0-or-later
"""Figure Settings Dialog."""
from gettext import gettext as _

from gi.repository import Graphs

from graphs import misc, styles, ui


class FigureSettingsDialog(Graphs.FigureSettingsDialog):
    """Figure Settings Dialog."""

    __gtype_name__ = "GraphsPythonFigureSettingsDialog"

    def __init__(
        self,
        application: Graphs.Application,
        highlighted: str = None,
    ):
        """Initialize the Figure Settings window and set the widget entries."""
        super().__init__(application=application)
        self.props.style_editor = styles.StyleEditor(self)
        self.setup(highlighted)
        self.connect("copy-request", self.copy_style)
        self.connect("closed", self.on_close)
        self.connect("entry-change", self.on_entry_change)
        self.connect("set-as-default", self.on_set_as_default)

    @staticmethod
    def on_entry_change(self, entry, prop) -> None:
        """Bind the entry upon change."""
        is_valid, value = ui.validate_entry(entry)
        if is_valid:
            self.props.figure_settings.set_property(prop, value)

    @staticmethod
    def copy_style(self, template, name) -> None:
        """Open the new style window."""
        style_manager = self.props.application.get_figure_style_manager()
        style_manager.copy_style(template, name)

    def on_close(self, *_args) -> None:
        """
        Handle closing.

        Closes the figure settings, saves the current style and adds the
        new state to the clipboard
        """
        self.emit("save-style-request")
        data = self.props.application.get_data()
        data.add_view_history_state()
        data.add_history_state()

    @staticmethod
    def on_set_as_default(self) -> None:
        """Set the current figure settings as the new default."""
        figure_settings = self.props.figure_settings
        settings = self.props.application.get_settings_child("figure")
        ignorelist = ["min_selected", "max_selected"] + misc.LIMITS
        for prop in dir(figure_settings.props):
            if prop in ignorelist:
                continue
            value = figure_settings.get_property(prop)
            prop = prop.replace("_", "-")
            if isinstance(value, str):
                settings.set_string(prop, value)
            elif isinstance(value, bool):
                settings.set_boolean(prop, value)
            elif isinstance(value, int):
                settings.set_enum(prop, value)
        self.add_toast_string(_("Defaults Updated"))



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for parsing files to usable data."""
import re
from gettext import gettext as _

from gi.repository import Gio, Graphs

from graphs import file_io, item, misc, project, utilities
from graphs.misc import ParseError

import numpy


def import_from_project(_params, _style, file: Gio.File) -> misc.ItemList:
    """Import data from project file."""
    project_dict = project.read_project_file(file)
    return list(map(item.new_from_dict, project_dict["data"]))


def import_from_xrdml(_params, style, file: Gio.File) -> misc.ItemList:
    """Import data from xrdml file."""
    content = file_io.parse_xml(file)
    intensities = content.getElementsByTagName("intensities")
    counting_time = content.getElementsByTagName("commonCountingTime")
    counting_time = float(counting_time[0].firstChild.data)
    ydata = intensities[0].firstChild.data.split()
    ydata = [int(value) / counting_time for value in ydata]

    scan_type = content.getElementsByTagName("scan")
    scan_axis = scan_type[0].attributes["scanAxis"].value
    if scan_axis.startswith("2Theta"):
        scan_axis = "2Theta"
    if scan_axis.startswith("Omega"):
        scan_axis = "Omega"

    data_points = content.getElementsByTagName("positions")
    for position in data_points:
        axis = position.attributes["axis"]
        if axis.value == scan_axis:
            unit = position.attributes["unit"].value
            start_pos = position.getElementsByTagName("startPosition")
            end_pos = position.getElementsByTagName("endPosition")
            start_pos = float(start_pos[0].firstChild.data)
            end_pos = float(end_pos[0].firstChild.data)
            xdata = numpy.linspace(start_pos, end_pos, len(ydata))
            xdata = numpy.ndarray.tolist(xdata)
    return [
        item.DataItem.new(
            style,
            xdata,
            ydata,
            name=Graphs.tools_get_filename(file),
            xlabel=f"{scan_axis} ({unit})",
            ylabel=_("Intensity (cps)"),
        ),
    ]


def import_from_xry(_params, style, file: Gio.File) -> misc.ItemList:
    """Import data from .xry files used by Leybold X-ray apparatus."""
    with file_io.open_wrapped(file, "rt", encoding="ISO-8859-1") as wrapper:

        def skip(lines: int):
            for _count in range(lines):
                next(wrapper)

        if wrapper.readline().strip() != "XR01":
            raise ParseError(_("Invalid .xry format"))
        skip(3)
        b_params = wrapper.readline().strip().split()
        x_step = float(b_params[3])
        x_value = float(b_params[0])

        skip(12)
        info = wrapper.readline().strip().split()
        item_count = int(info[0])

        name = Graphs.tools_get_filename(file)
        items = [
            item.DataItem.new(
                style,
                name=f"{name} - {i + 1}" if item_count > 1 else name,
                xlabel=_("β (°)"),
                ylabel=_("R (1/s)"),
            ) for i in range(item_count)
        ]
        for _count in range(int(info[1])):
            for value, item_ in zip(wrapper.readline().strip().split(), items):
                if value != "NaN":
                    item_.xdata.append(x_value)
                    item_.ydata.append(float(value))
            x_value += x_step
        skip(9 + item_count)
        for _count in range(int(wrapper.readline().strip())):
            values = wrapper.readline().strip().split()
            text = " ".join(values[7:])
            items.append(
                item.TextItem.new(
                    style,
                    float(values[5]),
                    float(values[6]),
                    text,
                    name=text,
                ),
            )
        return items


_PH = "dVldZaXqENhuPLPw"


def import_from_columns(params, style, file: Gio.File) -> misc.ItemList:
    """Import data from columns file."""
    item_ = item.DataItem.new(style, name=Graphs.tools_get_filename(file))
    column_x = params.get_int("column-x")
    column_y = params.get_int("column-y")
    separator = params.get_string("separator").replace(" ", "")
    skip_rows = params.get_int("skip-rows")
    delimiter = misc.DELIMITERS[params.get_string("delimiter")]
    if delimiter == "custom":
        delimiter = params.get_string("custom-delimiter")
    stream = Gio.DataInputStream.new(file.read(None))
    start_values = False
    for index, line in \
            enumerate(file_io.iter_data_stream(stream), -skip_rows):
        if index < 0:
            continue
        values = re.split(delimiter, line)
        if separator == ",":
            values = (
                string.replace(",", _PH).replace(".", ", ").replace(_PH, ".")
                for string in values
            )
        try:
            if len(values) == 1:
                float_value = utilities.string_to_float(values[0])
                if float_value is not None:
                    item_.ydata.append(float_value)
                    item_.xdata.append(index)
                    start_values = True
            else:
                try:
                    x_value = utilities.string_to_float(values[column_x])
                    y_value = utilities.string_to_float(values[column_y])
                    if x_value is None or y_value is None:
                        raise ValueError
                    else:
                        item_.xdata.append(x_value)
                        item_.ydata.append(y_value)
                        start_values = True
                except IndexError as error:
                    raise ParseError(
                        _("Import failed, column index out of range"),
                    ) from error
        # If not all values in the line are floats, start looking for
        # headers instead
        except ValueError:
            # Don't try to add headers when started adding values
            if not start_values:
                try:
                    headers = re.split(delimiter, line)
                    if len(values) == 1:
                        item_.set_ylabel(headers[column_x])
                    else:
                        item_.set_xlabel(headers[column_x])
                        item_.set_ylabel(headers[column_y])
                # If no label could be found at the index, skip.
                except IndexError:
                    pass
    stream.close()
    if not item_.xdata:
        raise ParseError(_("Unable to import from file"))
    return [item_]



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for file operations."""
import io
import json
from xml.dom import minidom

from gi.repository import GLib, Gio

from graphs import item


class FileLikeWrapper(io.BufferedIOBase):
    """FileLike Wrapper for Gio.Files."""

    def __init__(self, read_stream=None, write_stream=None):
        self._read_stream, self._write_stream = read_stream, write_stream

    @classmethod
    def new_for_io_stream(cls, io_stream: Gio.IOStream):
        """Create a wrapper for an IOStream."""
        return cls(
            read_stream=io_stream.get_input_stream(),
            write_stream=io_stream.get_output_stream(),
        )

    @property
    def closed(self) -> bool:
        """Whether or not the stream is closed."""
        return self._read_stream is None and self._write_stream is None

    def close(self) -> None:
        """Close the stream."""
        if self._read_stream is not None:
            self._read_stream.close()
            self._read_stream = None
        if self._write_stream is not None:
            self._write_stream.close()
            self._write_stream = None

    def writable(self) -> bool:
        """Whether or not the stream can be written to."""
        return self._write_stream is not None

    def write(self, b) -> int:
        """Write to the stream."""
        if self._write_stream is None:
            raise OSError()
        elif b is None or b == b"":
            return 0
        return self._write_stream.write_bytes(GLib.Bytes(b))

    def readable(self) -> bool:
        """Whether or not the stream can be read from."""
        return self._read_stream is not None

    def read(self, size=-1):
        """Read from the stream."""
        if self._read_stream is None:
            raise OSError()
        elif size == 0:
            return b""
        elif size > 0:
            return self._read_stream.read_bytes(size, None).get_data()
        buffer = io.BytesIO()
        while True:
            chunk = self._read_stream.read_bytes(4096, None)
            if chunk.get_size() == 0:
                break
            buffer.write(chunk.get_data())
        return buffer.getvalue()

    read1 = read


def create_write_stream(file: Gio.File) -> Gio.OutputStream:
    """Create a write stream for a given file."""
    if file.query_exists(None):
        file.delete(None)
    return file.create(0, None)


def open_wrapped(file: Gio.File, mode: str = "rt", encoding: str = "utf-8"):
    """Open a file in a FileLike wrapper."""
    read = "r" in mode
    append = "a" in mode
    replace = "w" in mode

    def _io_stream():
        return FileLikeWrapper.new_for_io_stream(file.open_readwrite(None))

    if "x" in mode:
        if file.query_exists():
            return OSError()
        stream = create_write_stream(file)
        stream.close()
    if read and append:
        obj = _io_stream()
    elif read and replace:
        stream = create_write_stream(file)
        stream.close()
        obj = _io_stream()
    elif read:
        obj = FileLikeWrapper(read_stream=file.read(None))
    elif replace:
        obj = FileLikeWrapper(write_stream=create_write_stream(file))
    elif append:
        obj = FileLikeWrapper(write_stream=file.append(None))

    if "b" not in mode:
        obj = io.TextIOWrapper(obj, encoding=encoding)
    return obj


def iter_data_stream(stream: Gio.DataInputStream):
    """
    Iterate over a data stream.

    Note: This can be removed in the next release of pygobject.
    """
    line = stream.read_line_utf8(None)[0]
    while line is not None:
        yield line
        line = stream.read_line_utf8(None)[0]


def save_item(file: Gio.File, item_: item.DataItem) -> None:
    """Save an Item to a txt file."""
    delimiter = "\t"
    fmt = delimiter.join(["%.12e"] * 2)
    xlabel, ylabel = item_.get_xlabel(), item_.get_ylabel()
    stream = Gio.DataOutputStream.new(create_write_stream(file))
    if xlabel != "" and ylabel != "":
        stream.stream(xlabel + delimiter + ylabel + "\n")
    for values in zip(item_.xdata, item_.ydata):
        stream.put_string(fmt % values + "\n")
    stream.close()


def parse_json(file: Gio.File) -> dict:
    """Parse a json file to a python dict."""
    with open_wrapped(file, "rb") as wrapper:
        return json.load(wrapper)


def write_json(file: Gio.File, json_object: dict, pretty_print=True) -> None:
    """Write a python dict to a python file."""
    with open_wrapped(file, "wt") as wrapper:
        json.dump(
            json_object,
            wrapper,
            indent=4 if pretty_print else None,
            sort_keys=True,
        )


def parse_xml(file: Gio.File) -> dict:
    """Parse a xml file to a python dict."""
    with open_wrapped(file, "rb") as wrapper:
        return minidom.parse(wrapper)



# SPDX-License-Identifier: GPL-3.0-or-later
"""UI representation of an Item."""
from gettext import gettext as _

from gi.repository import Adw, Gio, Graphs

from graphs.curve_fitting import CurveFittingDialog
from graphs.edit_item import EditItemDialog


class ItemBox(Graphs.ItemBox):
    """UI representation of an Item."""

    __gtype_name__ = "GraphsPythonItemBox"

    def __init__(self, application, item, index):
        super().__init__(
            application=application,
            item=item,
            index=index,
        )
        self.setup()
        self.props.drop_target.connect("drop", self.on_dnd_drop)

        action_list = ["edit", "delete", "curve_fitting"]
        if self.props.index > 0:
            action_list += ["move_up"]
        if self.props.index + 1 < len(self.props.application.get_data()):
            action_list += ["move_down"]
        action_group = Gio.SimpleActionGroup()

        # Set the action group for the widget
        for name in action_list:
            action = Gio.SimpleAction.new(name, None)
            action.connect("activate", getattr(self, f"{name}"))
            action_group.add_action(action)
        self.insert_action_group("item_box", action_group)

    def _change_position(self, source_index: int) -> None:
        data = self.props.application.get_data()
        data.change_position(self.props.index, source_index)
        data.add_history_state()
        data.add_view_history_state()

    def on_dnd_drop(self, _drop_target, value: int, _x, _y) -> None:
        """Handle dropped data."""
        self._change_position(value)

    def curve_fitting(self, _action, _shortcut) -> None:
        """Open Curve Fitting dialog."""
        CurveFittingDialog(self.props.application, self.props.item)

    def move_up(self, _action, _shortcut) -> None:
        """Move item up in hirarchy."""
        self._change_position(self.props.index - 1)

    def move_down(self, _action, _shortcut) -> None:
        """Move item down in hirarchy."""
        self._change_position(self.props.index + 1)

    def delete(self, _action, _shortcut) -> None:
        """Delete Item."""
        name = self.props.item.get_name()
        self.props.application.get_data().delete_items([self.props.item])
        toast = Adw.Toast.new(_("Deleted {name}").format(name=name))
        toast.set_button_label(_("Undo"))
        toast.set_action_name("app.undo")
        self.props.application.get_window().add_toast(toast)

    def edit(self, _action, _shortcut) -> None:
        """Show Edit Item Dialog."""
        EditItemDialog(self.props.application, self.props.item)



# SPDX-License-Identifier: GPL-3.0-or-later
"""Various utility functions."""
import ast
import contextlib
import operator as op
import re
from gettext import gettext as _
from numbers import Number

from gi.repository import Gio, Gtk

from graphs.misc import FUNCTIONS

import numpy

import sympy


def sig_fig_round(number: float, digits: int) -> float:
    """Round a number to the specified number of significant digits."""
    try:
        # Convert to scientific notation, and get power
        power = "{:e}".format(float(number)).split("e")[1]
    except IndexError:
        return None
    return round(float(number), -(int(power) - digits + 1))


def get_value_at_fraction(
    fraction: float,
    start: float,
    end: float,
    scale: int,
) -> float:
    """
    Get value at axis fraction.

    Obtain the selected value of an axis given at which percentage (in terms of
    fraction) of the length this axis is selected given the start and end range
    of this axis.
    """
    if scale == 0 or scale == 2:  # Linear or radian scale
        return start + fraction * (end - start)
    elif scale == 1:  # Logarithmic scale
        log_start = numpy.log10(start)
        log_end = numpy.log10(end)
        log_range = log_end - log_start
        log_value = log_start + log_range * fraction
        return pow(10, log_value)
    elif scale == 3:  # Square root scale
        # Use min limit as defined by scales.py
        start = max(0, start)
        sqrt_start = numpy.sqrt(start)
        sqrt_end = numpy.sqrt(end)
        sqrt_range = sqrt_end - sqrt_start
        sqrt_value = sqrt_start + sqrt_range * fraction
        return sqrt_value * sqrt_value
    elif scale == 4:  # Inverted scale (1/X)'
        # Use min limit as defined by scales.py if min equals zero
        start = end / 10 if end > 0 and start <= 0 else start
        scaled_range = 1 / start - 1 / end

        # Calculate the inverse-scaled value at the given percentage
        return 1 / (1 / end + fraction * scaled_range)


def get_fraction_at_value(
    value: float,
    start: float,
    end: float,
    scale: int,
) -> float:
    """
    Get fraction of axis at absolute value.

    Obtain the fraction of the total length of the selected axis a specific
    value corresponds to given the start and end range of the axis.
    """
    if scale == 0 or scale == 2:  # Linear or radian scale
        return (value - start) / (end - start)
    elif scale == 1:  # Logarithmic scale
        log_start = numpy.log10(start)
        log_end = numpy.log10(end)
        log_value = numpy.log10(value)
        log_range = log_end - log_start
        return (log_value - log_start) / log_range
    elif scale == 3:  # Square root scale
        # Use min limit as defined by scales.py
        start = max(0, start)
        sqrt_start = numpy.sqrt(start)
        sqrt_end = numpy.sqrt(end)
        sqrt_value = numpy.sqrt(value)
        sqrt_range = sqrt_end - sqrt_start
        return (sqrt_value - sqrt_start) / sqrt_range
    elif scale == 4:  # Inverted scale (1/X)
        # Use min limit as defined by scales.py if min equals zero
        start = end / 10 if end > 0 and start <= 0 else start
        scaled_range = 1 / start - 1 / end

        # Calculate the scaled percentage corresponding to the data point
        scaled_data_point = 1 / value
        return (scaled_data_point - 1 / end) / scaled_range


def create_file_filters(filters, add_all: bool = True) -> Gio.ListStore:
    """
    Create file filters.

    filters should be in the format:
    [
        (name, (suffix_a, suffix_b),
    ]
    """
    list_store = Gio.ListStore()
    for name, suffix_list in filters:
        file_filter = Gtk.FileFilter()
        file_filter.set_name(name)
        for suffix in suffix_list:
            file_filter.add_suffix(suffix)
        list_store.append(file_filter)
    if add_all:
        file_filter = Gtk.FileFilter()
        file_filter.set_name(_("All Files"))
        file_filter.add_pattern("*")
        list_store.append(file_filter)
    return list_store


def string_to_float(string: str) -> float:
    """Evaluate a string represantation of a number."""
    try:

        return _eval(ast.parse(preprocess(string.strip()), mode="eval").body)
    except (SyntaxError, ValueError):
        return None


OPERATORS = {
    ast.Add: op.add,
    ast.Sub: op.sub,
    ast.Mult: op.mul,
    ast.Div: op.truediv,
    ast.Pow: op.pow,
    ast.BitXor: op.xor,
    ast.USub: op.neg,
}


def _eval(node):
    # <number>
    if isinstance(node, ast.Constant) and isinstance(node.value, Number):
        return node.value
    elif isinstance(node, ast.BinOp):  # <left> <operator> <right>
        return OPERATORS[type(node.op)](_eval(node.left), _eval(node.right))
    elif isinstance(node, ast.UnaryOp):  # <operator> <operand> e.g., -1
        return OPERATORS[type(node.op)](_eval(node.operand))
    else:
        raise ValueError


def preprocess(string: str) -> str:
    """Preprocess an equation to be compatible with numexpr syntax."""

    def convert_degrees(match):
        """Convert degree expressions to radian expressions."""
        function, remainder = match.group(1), match.group(2)
        if function not in FUNCTIONS:
            return f"{function}{remainder}"
        expression, rest = _extract_expression(remainder)
        return f"{function}({expression}*{numpy.pi}/180){rest}"

    def _extract_expression(remainder):
        """Isolate the expression within the first pair of parentheses."""
        stack = []
        for i, char in enumerate(remainder.lower()):
            if char == "(":
                stack.append(char)
            elif char == ")":
                stack.pop()
                if not stack:  # Matching parenthesis found
                    stop_index = i + 1
                    break
        expression = remainder[:stop_index]
        rest = remainder[stop_index:]
        return expression, rest

    def convert_degrees_recursive(old_string):
        """Recursively convert degrees to match all parenthesis properly."""
        new_string = re.sub(r"(\w+)d(\(.*\))", convert_degrees, old_string)
        if new_string != old_string:
            return convert_degrees_recursive(new_string)
        return new_string

    def convert_cot(match):
        """Convert cotangent expressions to reciprocal tangent expressions."""
        expression = match.group(1)  # Get the content inside the brackets
        return f"1/(tan({expression}))"

    def convert_sec(match):
        """Convert secant expressions to reciprocal cosine expressions."""
        expression = match.group(1)  # Get the content inside the brackets
        return f"1/(cos({expression}))"

    def convert_csc(match):
        """Convert cosecant expressions to reciprocal sine expressions."""
        expression = match.group(1)  # Get the content inside the brackets
        return f"1/(sin({expression}))"

    def convert_arccot(match):
        """Convert arccotangent to reciprocal arcsine expressions."""
        expression = match.group(1)  # Get the content inside the brackets
        return f"arcsin(1/sqrt(1+{expression}**2))"

    def convert_arcsec(match):
        """Convert arcsecant to reciprocal cosine expressions."""
        expression = match.group(1)  # Get the content inside the brackets
        return f"(arccos(1/({expression})))"

    def convert_arccsc(match):
        """Convert arccosecant to reciprocal arcsine expressions."""
        expression = match.group(1)  # Get the content inside the brackets
        return f"(arcsin(1/({expression})))"

    def convert_superscript(match):
        """Convert superscript expressions to Python's power operator."""
        superscript_mapping = {
            "⁰": "0",
            "¹": "1",
            "²": "2",
            "³": "3",
            "⁴": "4",
            "⁵": "5",
            "⁶": "6",
            "⁷": "7",
            "⁸": "8",
            "⁹": "9",
        }
        sequence = match.group(1)  # Get the content inside the superscript
        sequence = "".join(
            superscript_mapping.get(char, char) for char in sequence
        )
        return f"**{sequence}"

    def add_asterix(match):
        """
        Add asterix to an equation.

        Adds asterix in equation when missing in case a number is followed
        by an alphabetical character, and adds parantheses around.

        Pattern is to check for least one digit, followed by at least one
        alphabetical character. e.g y = 24*x + 3sigma -> y = (24*x) + (3*sigma)
        """
        var, exp2 = match.group(1), match.group(2).lower()
        if var in FUNCTIONS:
            return f"{var}{exp2}"
        else:
            return f"{var}*{exp2}"

    string = string.lower()
    string = string.replace(",", ".")
    string = convert_degrees_recursive(string)
    string = re.sub(
        r"([\u2070-\u209f\u00b0-\u00be]+)",
        convert_superscript,
        string,
    )
    string = re.sub(r"(\d*\.?\d+)(?![Ee]?[-+]?\d)(\w+)", add_asterix, string)
    string = re.sub(r"(\w+)(\([\w\(]+)", add_asterix, string)
    string = string.replace("pi", f"({float(numpy.pi)})")
    string = string.replace("^", "**")
    string = string.replace(")(", ")*(")
    string = re.sub(r"arccot\((.*?)\)", convert_arccot, string)
    string = re.sub(r"arcsec\((.*?)\)", convert_arcsec, string)
    string = re.sub(r"arccsc\((.*?)\)", convert_arccsc, string)
    string = re.sub(r"cot\((.*?)\)", convert_cot, string)
    string = re.sub(r"sec\((.*?)\)", convert_sec, string)
    return re.sub(r"csc\((.*?)\)", convert_csc, string)


def string_to_function(equation_name: str) -> sympy.FunctionClass:
    """Convert a string into a sympy function."""
    variables = ["x"] + get_free_variables(equation_name)
    sym_vars = sympy.symbols(variables)
    with contextlib.suppress(sympy.SympifyError, TypeError, SyntaxError):
        symbolic = sympy.sympify(
            equation_name,
            locals=dict(zip(variables, sym_vars)),
        )
        return sympy.lambdify(sym_vars, symbolic)


def get_free_variables(equation_name: str) -> list:
    """Get the free variables (non-x) from an equation."""
    pattern = (
        r"\b(?!x\b|X\b"  # Exclude 'x' and 'X'
        r"|sec\b|sin\b|cos\b|log\b|tan\b|csc\b|cot\b"  # Exclude trig func.
        r"|arcsin\b|arccos\b|arctan\b"  # Exclude arctrig func.
        r"|arccot\b|arcsec\b|arccsc\b"  # Exclude arctrig func.
        r"|sinh\b|cosh\b|tanh\b"  # Exclude hyperbolicus argtrig func.
        r"|arcsinh\b|arccosh\b|arctanh\b"  # Exclude hyperb. arctrig func.
        r"|exp\b|sqrt\b|abs\b|log10\b)"  # Exclude 'exp', 'sqrt', 'abs'
        r"[a-zA-Z]+\b"  # Match any character sequence that is not excluded
    )
    return list(set(re.findall(pattern, equation_name)))



# SPDX-License-Identifier: GPL-3.0-or-later
"""Data management module."""
import copy
import math
import os
from gettext import gettext as _
from urllib.parse import unquote, urlparse

from gi.repository import GObject, Gio, Graphs

from graphs import item, misc

import numpy

_FIGURE_SETTINGS_HISTORY_IGNORELIST = misc.LIMITS + [
    "min_selected",
    "max_selected",
]


class Data(Graphs.Data):
    """Class for managing data."""

    __gtype_name__ = "GraphsPythonData"

    def __init__(self, settings: Gio.Settings):
        figure_settings = Graphs.FigureSettings.new(settings)
        super().__init__(
            settings=settings,
            figure_settings=figure_settings,
        )
        self._initialize()
        figure_settings.connect("notify", self._on_figure_settings_change)
        self.connect("notify::unsaved", self._on_unsaved_change)
        self._update_used_positions()
        self.connect(
            "add_history_state_request",
            lambda _s: self.add_history_state(),
        )

    def reset(self):
        """Reset data."""
        # Reset figure settings
        default_figure_settings = Graphs.FigureSettings.new(
            self.props.settings,
        )
        figure_settings = self.get_figure_settings()
        for prop in dir(default_figure_settings.props):
            new_value = default_figure_settings.get_property(prop)
            figure_settings.set_property(prop, new_value)

        # Reset items
        if not self.empty:
            self.delete_items([item for item in self])
        self._initialize()
        self.props.file = None
        self.props.unsaved = False

    def _initialize(self):
        """Initialize the data class and set default values."""
        figure_settings = self.get_figure_settings()
        limits = figure_settings.get_limits()
        self.props.can_undo = False
        self.props.can_redo = False
        self.props.can_view_back = False
        self.props.can_view_forward = False
        self._history_states = [([], limits)]
        self._history_pos = -1
        self._view_history_states = [limits]
        self._view_history_pos = -1
        self._items = {}
        self._set_data_copy()

    @GObject.Property(type=bool, default=True, flags=1 | 1073741824)
    def empty(self) -> bool:
        """Whether or not the class is empty."""
        return not self._items

    def _on_unsaved_change(self, _a, _b) -> None:
        if not self.props.unsaved:
            self.emit("saved")
        self.notify("project-name")
        self.notify("project-path")

    @GObject.Property(type=str, default="", flags=1)
    def project_name(self) -> str:
        """Get project_name property."""
        if self.props.file is None:
            title = _("Untitled Project")
        else:
            title = Graphs.tools_get_filename(self.props.file)
        if self.props.unsaved:
            title = "• " + title
        return title

    @GObject.Property(type=str, default="", flags=1)
    def project_path(self) -> str:
        """Retrieve the path of the associated file."""
        if self.props.file is None:
            return _("Draft")
        uri_parse = urlparse(self.props.file.get_uri())
        filepath = os.path.dirname(
            os.path.join(uri_parse.netloc, unquote(uri_parse.path)),
        )
        if filepath.startswith("/var"):
            # Fix for rpm-ostree distros, where home is placed in /var/home
            filepath = filepath.replace("/var", "", 1)
        return filepath.replace(os.path.expanduser("~"), "~")

    @GObject.Property(type=bool, default=False, flags=1)
    def items_selected(self) -> bool:
        """Whether or not at least one item is selected."""
        return any(item_.get_selected() for item_ in self)

    @GObject.Property(type=object, flags=3 | 1073741824)  # explicit notify
    def items(self) -> misc.ItemList:
        """All managed items."""
        return self.get_items()

    @items.setter
    def items(self, items: misc.ItemList) -> None:
        """Set items property."""
        self.set_items(items)

    def get_items(self) -> misc.ItemList:
        """Get all managed items."""
        return list(self._items.values())

    def set_items(self, items: misc.ItemList) -> None:
        """Set all managed items."""
        self._items = {}
        for item_ in items:
            self._add_item(item_)
        self._update_used_positions()
        self.notify("items")
        self.notify("empty")

    def _add_item(self, item_: Graphs.Item) -> None:
        """Append items to self."""
        self._connect_to_item(item_)
        self._items[item_.get_uuid()] = item_

    def _delete_item(self, uuid: str) -> None:
        """Pop and delete item."""
        item_ = self._items[uuid]
        self._items.pop(uuid)
        del item_

    def index(self, item_: Graphs.Item) -> int:
        """Get the index of an item."""
        return list(self._items.keys()).index(item_.get_uuid())

    def get_names(self) -> list[str]:
        """All items' names."""
        return [item_.get_name() for item_ in self]

    def get_n_items(self) -> int:
        """Amount of managed items."""
        return len(self._items)

    def get_at_pos(self, position: int) -> Graphs.Item:
        """Get item at position."""
        return self.get_items()[position]

    def get_for_uuid(self, uuid: str) -> Graphs.Item:
        """Get item for key."""
        return self._items[uuid]

    def __len__(self) -> int:
        """Magic alias for `get_n_items()`."""
        return self.get_n_items()

    def __iter__(self):
        """Iterate over items."""
        return iter(self._items.values())

    def __getitem__(self, getter: str | int):
        """Magic alias for retrieving items."""
        if isinstance(getter, str):
            return self.get_for_uuid(getter)
        return self.get_at_pos(getter)

    def change_position(self, index1: int, index2: int) -> None:
        """Change item position of index2 to that of index1."""
        items = self.get_items()
        # Check if target key is lower in the order, if so we can put the old
        # key below the target key. Otherwise put it above.
        if index1 < index2:
            items[index1:index2 + 1] = [items[index2]] + items[index1:index2]
        else:
            items[index2:index1 + 1] = \
                items[index2 + 1:index1 + 1] + [items[index2]]
        self.set_items(items)
        self._current_batch.append((3, (index2, index1)))

    def add_items(
        self,
        items: misc.ItemList,
        style_manager: Graphs.StyleManager,
    ) -> None:
        """
        Add items to be managed.

        Respects settings in regards to handling duplicate names.
        New Items with a x- or y-label change the figures current labels if
        they are still the default. If they are already modified and do not
        match the items label, they get moved to another axis.
        """
        figure_settings = self.get_figure_settings()
        selected_style = style_manager.get_selected_style_params()
        color_cycle = selected_style["axes.prop_cycle"].by_key()["color"]
        used_colors = []

        def _append_used_color(color):
            used_colors.append(color)
            if len(set(used_colors)) == len(color_cycle):
                for color in color_cycle:
                    used_colors.remove(color)

        def _is_default(prop):
            return figure_settings.get_property(prop) == \
                self.props.settings.get_string(prop)

        for item_ in self:
            color = item_.get_color()
            if color in color_cycle:
                _append_used_color(color)
        used_names = set(self.get_names())
        for new_item in items:
            item_name = new_item.get_name()
            if item_name in used_names:
                new_item.set_name(
                    Graphs.tools_get_duplicate_string(
                        item_name,
                        list(used_names),
                    ),
                )
            used_names.add(new_item.get_name())
            xlabel = new_item.get_xlabel()
            if xlabel:
                original_position = new_item.get_xposition()
                if original_position == 0:
                    if _is_default("bottom-label") or self.props.empty:
                        figure_settings.set_bottom_label(xlabel)
                    elif xlabel != figure_settings.get_bottom_label():
                        new_item.set_xposition(1)
                if new_item.get_xposition() == 1:
                    if _is_default("top-label"):
                        figure_settings.set_top_label(xlabel)
                    elif xlabel != figure_settings.get_top_label():
                        new_item.set_xposition(original_position)
            ylabel = new_item.get_ylabel()
            if ylabel:
                original_position = new_item.get_yposition()
                if original_position == 0:
                    if _is_default("left-label") or self.props.empty:
                        figure_settings.set_left_label(ylabel)
                    elif ylabel != figure_settings.get_left_label():
                        new_item.set_yposition(1)
                if new_item.get_yposition() == 1:
                    if _is_default("right-label"):
                        figure_settings.set_right_label(ylabel)
                    elif ylabel != figure_settings.get_right_label():
                        new_item.set_yposition(original_position)
            if new_item.get_color() == "":
                for color in color_cycle:
                    if color not in used_colors:
                        _append_used_color(color)
                        new_item.set_color(color)
                        break

            self._add_item(new_item)
            change = (1, copy.deepcopy(new_item.to_dict()))
            self._current_batch.append(change)
        self.optimize_limits()
        self.add_history_state()
        self._update_used_positions()
        self.notify("items")
        self.notify("empty")
        self.notify("items_selected")

    def delete_items(self, items: misc.ItemList):
        """Delete specified items."""
        settings = self.get_figure_settings()
        for item_ in items:
            self._current_batch.append(
                (2, (self.index(item_), item_.to_dict())),
            )
            x_position = item_.get_xposition()
            y_position = item_.get_yposition() + 2
            xlabel = item_.get_xlabel()
            ylabel = item_.get_ylabel()
            self._delete_item(item_.get_uuid())
        used = self.get_used_positions()
        for position in [x_position, y_position]:
            direction = misc.DIRECTIONS[position]
            item_label = xlabel if position < 2 else ylabel
            axis_label = getattr(settings, f"get_{direction}_label")()
            if not used[position] and item_label == axis_label:
                set_label = getattr(settings, f"set_{direction}_label")
                set_label(
                    self.props.settings.get_string(f"{direction}-label"),
                )

        self._update_used_positions()
        self.notify("items")
        self.notify("empty")
        self.add_history_state()
        self.notify("items_selected")

    def _connect_to_item(self, item_: Graphs.Item):
        item_.connect("notify::selected", self._on_item_select)
        item_.connect("notify", self._on_item_change)
        for prop in ("xposition", "yposition"):
            item_.connect(f"notify::{prop}", self._on_item_position_change)

    def _on_item_position_change(self, _item, _ignored) -> None:
        self.optimize_limits()
        self._update_used_positions()
        self.notify("items")

    def _on_item_select(self, _x, _y) -> None:
        self.notify("items_selected")

    def _on_item_change(self, item_, param) -> None:
        self._current_batch.append((
            0,
            (
                item_.get_uuid(),
                param.name,
                copy.deepcopy(self._data_copy[item_.get_uuid()][param.name]),
                copy.deepcopy(item_.get_property(param.name)),
            ),
        ))

    def _update_used_positions(self) -> None:
        # bottom, top, left, right
        figure_settings = self.get_figure_settings()
        used_positions = [False, False, False, False]

        for item_ in self.items:
            if (
                figure_settings.get_hide_unselected()
                and not item_.get_selected()
            ):
                continue
            used_positions[item_.get_xposition()] = True
            used_positions[item_.get_yposition() + 2] = True
        if not any(used_positions):
            self.set_used_positions(True, False, True, False)
            return
        self.set_used_positions(*used_positions)

    def _on_figure_settings_change(self, figure_settings, param) -> None:
        if param.name in _FIGURE_SETTINGS_HISTORY_IGNORELIST:
            return
        self._current_batch.append((
            4,
            (
                param.name,
                copy.deepcopy(self._figure_settings_copy[param.name]),
                copy.deepcopy(figure_settings.get_property(param.name)),
            ),
        ))

    def _set_data_copy(self) -> None:
        """Set a deep copy for the data."""
        self._current_batch: list = []
        self._data_copy = copy.deepcopy({
            item_.get_uuid(): item_.to_dict()
            for item_ in self
        })
        self._figure_settings_copy = copy.deepcopy({
            prop.replace("_", "-"):
            self.props.figure_settings.get_property(prop)
            for prop in dir(self.props.figure_settings.props)
        })

    def add_history_state(self, old_limits: misc.Limits = None) -> None:
        """Add a state to the clipboard."""
        if not self._current_batch:
            return
        if self._history_pos != -1:
            self._history_states = self._history_states[:self._history_pos + 1]
        self._history_pos = -1
        self._history_states.append(
            (self._current_batch, self.get_figure_settings().get_limits()),
        )
        if old_limits is not None:
            old_state = self._history_states[-2][1]
            for index in range(8):
                old_state[index] = old_limits[index]
        self.props.can_redo = False
        self.props.can_undo = True
        # Keep history states length limited to 100 spots
        if len(self._history_states) > 101:
            self._history_states = self._history_states[1:]
        self._set_data_copy()
        self.props.unsaved = True

    def undo(self) -> None:
        """Undo the latest change that was added to the clipboard."""
        if not self.props.can_undo:
            return
        batch = self._history_states[self._history_pos][0]
        self._history_pos -= 1
        items_changed = False
        for change_type, change in reversed(batch):
            if change_type == 0:
                self[change[0]].set_property(change[1], change[2])
            elif change_type == 1:
                self._delete_item(change["uuid"])
                items_changed = True
            elif change_type == 2:
                item_ = item.new_from_dict(copy.deepcopy(change[1]))
                self._add_item(item_)
                self.change_position(change[0], len(self) - 1)
                items_changed = True
            elif change_type == 3:
                self.change_position(change[0], change[1])
                items_changed = True
            elif change_type == 4:
                self.props.figure_settings.set_property(
                    change[0],
                    change[1],
                )
        if items_changed:
            self._update_used_positions()
            self.notify("items")
            self.notify("empty")
        self.notify("items_selected")
        self.get_figure_settings().set_limits(
            self._history_states[self._history_pos][1],
        )
        self.props.can_redo = True
        self.props.can_undo = \
            abs(self._history_pos) < len(self._history_states)
        self._set_data_copy()
        self.add_view_history_state()

    def redo(self) -> None:
        """Redo the latest change that was added to the clipboard."""
        if not self.props.can_redo:
            return
        self._history_pos += 1
        state = self._history_states[self._history_pos]
        items_changed = False
        for change_type, change in state[0]:
            if change_type == 0:
                self[change[0]].set_property(change[1], change[3])
            elif change_type == 1:
                self._add_item(item.new_from_dict(copy.deepcopy(change)))
                items_changed = True
            elif change_type == 2:
                self._delete_item(change[1]["uuid"])
                items_changed = True
            elif change_type == 3:
                self.change_position(change[1], change[0])
                items_changed = True
            elif change_type == 4:
                self.props.figure_settings.set_property(
                    change[0],
                    change[2],
                )
        if items_changed:
            self._update_used_positions()
            self.notify("items")
            self.notify("empty")
        self.notify("items_selected")
        self.get_figure_settings().set_limits(state[1])
        self.props.can_redo = self._history_pos < -1
        self.props.can_undo = True
        self._set_data_copy()
        self.add_view_history_state()

    def add_view_history_state(self) -> None:
        """Add the view to the view history."""
        limits = self.get_figure_settings().get_limits()
        if all(
            math.isclose(old, new) for old,
            new in zip(self._view_history_states[-1], limits)
        ):
            return
        # If a couple of redo's were performed previously, it deletes the
        # clipboard data that is located after the current clipboard
        # position and disables the redo button
        if self._view_history_pos != -1:
            self._view_history_states = \
                self._view_history_states[:self._view_history_pos + 1]
        self._view_history_pos = -1
        self._view_history_states.append(limits)
        self.props.can_view_back = True
        self.props.can_view_forward = False
        self.props.unsaved = True

    def view_back(self) -> None:
        """Move the view to the previous value in the view history."""
        if not self.props.can_view_back:
            return
        self._view_history_pos -= 1
        self.get_figure_settings().set_limits(
            self._view_history_states[self._view_history_pos],
        )
        self.props.can_view_forward = True
        self.props.can_view_back = \
            abs(self._view_history_pos) < len(self._view_history_states)

    def view_forward(self) -> None:
        """Move the view to the next value in the view history."""
        if not self.props.can_view_forward:
            return
        self._view_history_pos += 1
        self.get_figure_settings().set_limits(
            self._view_history_states[self._view_history_pos],
        )
        self.props.can_view_back = True
        self.props.can_view_forward = self._view_history_pos < -1

    def optimize_limits(self) -> None:
        """Optimize the limits of the canvas to the data class."""
        figure_settings = self.get_figure_settings()
        axes = [[
            direction,
            False,
            [],
            [],
            figure_settings.get_property(f"{direction}_scale"),
        ] for direction in ("bottom", "left", "top", "right")]
        for item_ in self:
            if not isinstance(item_, item.DataItem) or (
                not item_.get_selected()
                and figure_settings.get_hide_unselected()
            ):
                continue
            for index in \
                    item_.get_xposition() * 2, 1 + item_.get_yposition() * 2:
                axes[index][1] = True
                xydata = numpy.asarray(
                    item_.ydata if index % 2 else item_.xdata,
                )
                try:
                    xydata = xydata[numpy.isfinite(xydata)]
                except TypeError:
                    return
                nonzero_data = numpy.array([
                    value for value in xydata if value != 0
                ])
                axes[index][2].append(
                    nonzero_data.min() if axes[index][4] in (1, 4)
                    and len(nonzero_data) > 0 else xydata.min(),
                )
                axes[index][3].append(xydata.max())

        for count, (direction, used, min_all, max_all, scale) in \
                enumerate(axes):
            if not used:
                continue
            min_all = min(min_all)
            max_all = max(max_all)
            if scale != 1:  # For non-logarithmic scales
                span = max_all - min_all
                # 0.05 padding on y-axis, 0.015 padding on x-axis
                padding_factor = 0.05 if count % 2 else 0.015
                max_all += padding_factor * span

                # For inverse scale, calculate padding using a factor
                min_all = (
                    min_all - padding_factor * span if scale != 4 else min_all
                    * 0.99
                )
            else:  # Use different scaling type for logarithmic scale
                # Use padding factor of 2 for y-axis, 1.025 for x-axis
                padding_factor = 2 if count % 2 else 1.025
                min_all *= 1 / padding_factor
                max_all *= padding_factor
            figure_settings.set_property(f"min_{direction}", min_all)
            figure_settings.set_property(f"max_{direction}", max_all)
        self.add_view_history_state()

    def get_project_dict(self, version: str) -> dict:
        """Convert data to dict."""
        figure_settings = self.get_figure_settings()
        return {
            "version": version,
            "data": [item_.to_dict() for item_ in self],
            "figure-settings": {
                key: figure_settings.get_property(key)
                for key in dir(figure_settings.props)
            },
            "history-states": self._history_states,
            "history-position": self._history_pos,
            "view-history-states": self._view_history_states,
            "view-history-position": self._view_history_pos,
        }

    def load_from_project_dict(self, project_dict: dict) -> None:
        """Load data from dict."""
        # Load data
        figure_settings = self.get_figure_settings()
        for key, value in project_dict["figure-settings"].items():
            if figure_settings.get_property(key) != value:
                figure_settings.set_property(key, value)
        self.set_items(item.new_from_dict(d) for d in project_dict["data"])
        self.notify("items_selected")

        # Set clipboard
        self._set_data_copy()
        self._history_states = project_dict["history-states"]
        self._history_pos = project_dict["history-position"]
        self._view_history_states = project_dict["view-history-states"]
        self._view_history_pos = project_dict["view-history-position"]
        self.unsaved = False

        # Set clipboard/view buttons
        self.props.can_undo = \
            abs(self._history_pos) < len(self._history_states)
        self.props.can_redo = self._history_pos < -1
        self.props.can_view_back = \
            abs(self._view_history_pos) < len(self._view_history_states)
        self.props.can_view_forward = self._view_history_pos < -1



# SPDX-License-Identifier: GPL-3.0-or-later
"""Module for migrating old data to new structures."""
import contextlib
import pickle
import sys
from pathlib import Path

from gi.repository import GLib, Gio, Graphs

from graphs import file_io, misc, style_io

_CONFIG_MIGRATION_TABLE = {
    # old-key: (category, key, old-default)
    "action_center_data": ("actions", "center", "Center at middle coordinate"),
    "addequation_equation": ("add-equation", "equation", "X"),
    "addequation_step_size": ("add-equation", "step-size", "0.01"),
    "addequation_x_start": ("add-equation", "x-start", "0"),
    "addequation_x_stop": ("add-equation", "x-stop", "10"),
    "export_figure_dpi": ("export-figure", "dpi", 100),
    "export_figure_transparent": ("export-figure", "transparent", True),
    "hide_unselected": ("figure", "hide-unselected", False),
    "plot_custom_style": ("figure", "custom-style", "adwaita"),
    "plot_legend": ("figure", "legend", True),
    "plot_right_label": ("figure", "right-label", "Y Value 2"),
    "plot_right_scale": ("figure", "right-scale", "linear"),
    "plot_title": ("figure", "title", ""),
    "plot_top_label": ("figure", "top-label", "X Value 2"),
    "plot_top_scale": ("figure", "top-scale", "linear"),
    "plot_use_custom_style": ("figure", "use-custom-style", False),
    "plot_x_label": ("figure", "bottom-label", "X Value"),
    "plot_x_scale": ("figure", "bottom-scale", "linear"),
    "plot_y_label": ("figure", "left-label", "Y Value"),
    "plot_y_scale": ("figure", "left-scale", "linear"),
}

_CENTER_ACTION_MIGRATION_TABLE = {
    "Center at middle coordinate": "middle-x",
    "Center at middle X value": "max-y",
}


def migrate_config(settings: Gio.Settings) -> None:
    """Migrate old file-based user config to dconf."""
    main_dir = Gio.File.new_for_path(GLib.get_user_config_dir())
    old_config_dir = main_dir.get_child_for_display_name("Graphs")
    if not old_config_dir.query_exists(None):
        return
    new_config_dir = Graphs.tools_get_config_directory()
    config_file = old_config_dir.get_child_for_display_name("config.json")
    import_file = old_config_dir.get_child_for_display_name("import.json")
    old_styles_dir = old_config_dir.get_child_for_display_name("styles")
    if config_file.query_exists(None):
        _migrate_config(settings, config_file)
    if import_file.query_exists(None):
        _migrate_import_params(settings, import_file)
    if old_styles_dir.query_exists(None):
        _migrate_styles(old_styles_dir, new_config_dir)
    old_config_dir.delete(None)


def _migrate_config(settings_, config_file):
    config = file_io.parse_json(config_file)
    for old_key, (category, key, old_default) \
            in _CONFIG_MIGRATION_TABLE.items():
        with contextlib.suppress(KeyError, ValueError):
            value = config[old_key]
            if old_key == "action_center_data":
                value = _CENTER_ACTION_MIGRATION_TABLE[value]
            elif "scale" in key:
                value = value.capitalize()
            if old_default != value:
                settings_.get_child(category)[key] = value
    config_file.delete(None)


def _migrate_import_params(settings_, import_file):
    params = file_io.parse_json(import_file)["columns"]
    settings = settings_.get_child("import-params").get_child("columns")
    settings.set_string("separator", params["separator"] + " ")

    old_delimiter = params["delimiter"]
    if old_delimiter in misc.DELIMITERS.values():
        inverted_dict = {value: key for key, value in misc.DELIMITERS.items()}
        settings.set_string("delimiter", inverted_dict[old_delimiter])
    else:
        settings.set_string("delimiter", "custom")
        settings.set_string("custom-delimiter", old_delimiter)

    for key in ("column_x", "column_y", "skip_rows"):
        settings.set_int(key.replace("_", "-"), params[key])
    import_file.delete(None)


SYSTEM_STYLES = [
    "adwaita",
    "adwaita-dark",
    "bmh",
    "classic",
    "dark-background",
    "fivethirtyeight",
    "ggplot",
    "grayscale",
    "seaborn",
    "seaborn-bright",
    "seaborn-colorblind",
    "seaborn-dark",
    "seaborn-darkgrid",
    "seaborn-dark-pallete",
    "seaborn-deep",
    "seaborn-muted",
    "seaborn-notebook",
    "seaborn-paper",
    "seaborn-pastel",
    "seaborn-poster",
    "seaborn-talk",
    "seaborn-ticks",
    "seaborn-white",
    "seaborn-whitegrid",
    "solarized-light",
    "tableu-colorblind10",
    "thesis",
    "yaru",
    "yaru-dark",
]


def _migrate_styles(old_styles_dir, new_config_dir):
    new_styles_dir = new_config_dir.get_child_for_display_name("styles")
    if not new_styles_dir.query_exists(None):
        new_styles_dir.make_directory_with_parents()
    enumerator = old_styles_dir.enumerate_children("default::*", 0, None)
    adwaita = style_io.parse(
        Gio.File.new_for_uri(
            "resource:///se/sjoerd/Graphs/styles/adwaita.mplstyle",
        ),
    )[0]
    for file in map(enumerator.get_child, enumerator):
        stylename = Path(Graphs.tools_get_filename(file)).stem
        if stylename not in SYSTEM_STYLES:
            params = style_io.parse(file)[0]
            for key, value in adwaita.items():
                if key not in params:
                    params[key] = value
            style_io.write(
                new_styles_dir.get_child_for_display_name(
                    f"{stylename.lower().replace(' ', '-')}.mplstyle",
                ),
                stylename,
                params,
            )
        file.delete(None)
    enumerator.close(None)
    old_styles_dir.delete(None)


ITEM_MIGRATION_TABLE = {
    "plot_x_position": "xposition",
    "plot_y_position": "yposition",
    "x_anchor": "xanchor",
    "y_anchor": "yanchor",
    "key": "uuid",
}

ITEM_VALUE_MIGRATION_TABLE = {
    "linestyle": ["none", "solid", "dotted", "dashed", "dashdot"],
    "markerstyle": [
        "none",
        ".",
        ",",
        "o",
        "v",
        "^",
        "<",
        ">",
        "8",
        "s",
        "p",
        "*",
        "h",
        "H",
        "+",
        "x",
        "D",
        "d",
        "|",
        "_",
        "P",
        "X",
    ],
    "xposition": ["bottom", "top"],
    "yposition": ["left", "right"],
}

PLOT_SETTINGS_MIGRATION_TABLE = {
    "xlabel": "bottom_label",
    "ylabel": "left_label",
    "xscale": "bottom_scale",
    "yscale": "left_scale",
    "use_custom_plot_style": "use_custom_style",
    "custom_plot_style": "custom_style",
    "mix_right": "min_right",
}

LEGEND_POSITIONS = [
    "best",
    "upper right",
    "upper left",
    "lower left",
    "lower right",
    "center left",
    "center right",
    "lower center",
    "upper center",
    "center",
]


class PlotSettings:
    """Old PlotSettings standin."""

    def migrate(self) -> dict:
        """Migrate class to dict."""
        dictionary = {}
        for key, value in self.__dict__.items():
            with contextlib.suppress(KeyError):
                key = PLOT_SETTINGS_MIGRATION_TABLE[key]
            if "scale" in key:
                value = 0 if value == "linear" else 1
            elif key == "legend_position":
                value = LEGEND_POSITIONS.index(value)
            dictionary[key] = value
        return dictionary


class ItemBase:
    """Old ItemBase standin."""

    def migrate(self) -> dict:
        """Migrate class to dict."""
        dictionary = {"type": self.item_type}
        for key, value in self.__dict__.items():
            with contextlib.suppress(KeyError):
                key = ITEM_MIGRATION_TABLE[key]
            if key in ITEM_VALUE_MIGRATION_TABLE:
                try:
                    value = ITEM_VALUE_MIGRATION_TABLE[key].index(value)
                except IndexError:
                    value = 0
            dictionary[key] = value
        return dictionary


class Item(ItemBase):
    """Old Item standin."""

    item_type = "GraphsDataItem"


class TextItem(ItemBase):
    """Old TextItem standin."""

    item_type = "GraphsTextItem"


_DEFAULT_VIEW = [0, 1, 0, 10, 0, 1, 0, 10]


def migrate_project(file: Gio.File) -> dict:
    """Migrate pickle-based project."""
    sys.modules["graphs.misc"] = sys.modules[__name__]
    sys.modules["graphs.item"] = sys.modules[__name__]
    with file_io.open_wrapped(file, "rb") as wrapper:
        project = pickle.load(wrapper)

    figure_settings = project["plot_settings"].migrate()
    current_limits = [figure_settings[key] for key in misc.LIMITS]
    history_pos = project["clipboard_pos"]
    history_states = _migrate_clipboard(
        project["datadict_clipboard"],
        history_pos,
        current_limits,
    )

    return {
        "version": project["version"],
        "data": [item.migrate() for item in project["data"].values()],
        "figure-settings": figure_settings,
        "history-states": history_states,
        "history-position": history_pos,
        "view-history-states": [_DEFAULT_VIEW.copy(), current_limits],
        "view-history-position": -1,
    }


def _migrate_clipboard(clipboard, clipboard_pos, current_limits):
    if not clipboard:
        return []
    new_clipboard = []
    if len(clipboard) > 100:
        clipboard = clipboard[len(clipboard) - 100:]
    states = [{
        item.key: item.migrate()
        for item in state.values()
    } for state in clipboard]
    new_clipboard.append(([], _DEFAULT_VIEW.copy()))
    initial_items = states[1].values()
    new_clipboard.append((
        [(1, item) for item in initial_items],
        _get_limits(initial_items),
    ))
    if len(states) > 2:
        for count in range(len(states) - 2):
            batch = []
            previous_state = states[count + 1]
            current_state = states[count + 2]
            if len(current_state) < len(previous_state):
                for key, item in previous_state.copy().items():
                    if key not in current_state:
                        batch.append((2, previous_state.index(item), item))
                        previous_state.pop(item)
            for count_2, (key, item) in enumerate(current_state.items()):
                if key in previous_state:
                    previous_index = list(previous_state.keys()).index(key)
                    if previous_index != count_2:
                        batch.append((3, (previous_index, count_2)))
                    else:
                        for key_2, value in item.items():
                            previous_value = previous_state[key][key_2]
                            if value != previous_value:
                                batch.append(
                                    (0, (key, key_2, previous_value, value)),
                                )
                else:
                    batch.append((1, item))
            if clipboard_pos == count - len(states) + 1:
                limits = _get_limits(current_state.values())
            else:
                limits = current_limits
            new_clipboard.append((batch, limits))
    return new_clipboard


def _get_limits(items):
    limits = [None] * 8
    for item in items:
        if item["type"] != "Item":
            continue
        for count, x_or_y in enumerate(["x", "y"]):
            index = item[f"{x_or_y}position"] * 2 + 4 * count
            data = item[f"{x_or_y}data"]
            try:
                limits[index] = min(limits[index], min(data))
            except TypeError:
                limits[index] = min(data)
            try:
                limits[index + 1] = max(limits[index + 1], max(data))
            except TypeError:
                limits[index + 1] = max(data)
    for count in range(8):
        default_view_copy = _DEFAULT_VIEW.copy()
        if limits[count] is None:
            limits[count] = default_view_copy[count]
    return limits



# SPDX-License-Identifier: GPL-3.0-or-later
"""Curve fitting module."""
import re
from gettext import gettext as _

from gi.repository import Adw, Graphs, Gtk

from graphs import utilities
from graphs.canvas import Canvas
from graphs.item import DataItem, FillItem

import numpy

from scipy.optimize import _minpack, curve_fit


class CurveFittingDialog(Graphs.CurveFittingDialog):
    """Class for displaying the Curve Fitting dialog."""

    __gtype_name__ = "GraphsPythonCurveFittingDialog"

    def __init__(self, application: Graphs.Application, item: Graphs.Item):
        """Initialize the curve fitting dialog."""
        super().__init__(application=application)
        Adw.StyleManager.get_default().connect("notify", self.reload_canvas)
        self.connect("equation_change", self.on_equation_change)
        self.connect("fit_curve_request", self.fit_curve)
        self.connect("add_fit_request", self.add_fit)
        self.fitting_parameters = FittingParameterContainer()
        style = \
            application.get_figure_style_manager().get_system_style_params()

        self.param = []
        self.sigma = []
        self.r2 = 0

        # Generate items for the canvas
        self.data_curve = DataItem.new(
            style,
            xdata=item.xdata,
            ydata=item.ydata,
            name=item.get_name(),
            color="#1A5FB4",
        )
        self.data_curve.linestyle = 0
        self.data_curve.markerstyle = 1
        self.data_curve.markersize = 13
        self.fitted_curve = DataItem.new(style, color="#A51D2D")
        self.fill = FillItem.new(
            style,
            (
                self.fitted_curve.xdata,
                self.fitted_curve.ydata,
                self.fitted_curve.ydata,
            ),
            color="#1A5FB4",
            alpha=0.15,
        )

        self.reload_canvas()
        self.setup()
        self.present(application.get_window())

    def reload_canvas(self, *_args) -> None:
        """Reinitialise the currently used canvas."""
        application = self.props.application
        figure_settings = application.get_data().get_figure_settings()
        style = \
            application.get_figure_style_manager().get_system_style_params()
        canvas = Canvas(style, interactive=False)
        canvas.props.items = [self.fitted_curve, self.data_curve, self.fill]
        axis = canvas.axes[0]
        axis.yscale = "linear"
        axis.xscale = "linear"
        axis.xlabel = "X Value"
        canvas.left_label = figure_settings.get_property("left_label")
        canvas.bottom_label = figure_settings.get_property("bottom_label")
        self.set_canvas(canvas)

    @staticmethod
    def on_equation_change(self, equation: str) -> None:
        """
        Logic to execute whenever the equation is changed.

        Set the free variables and corresponding entry rows when the equation
        has been changed.
        """
        processed_equation = utilities.preprocess(equation)
        self.set_equation_string(processed_equation)
        free_variables = utilities.get_free_variables(processed_equation)
        if len(free_variables) == 0:
            self.set_results(error="equation")
            return
        self.fitting_parameters.update(free_variables)
        fit = self.fit_curve()
        if not fit:
            return False
        box = self.get_fitting_params_box()
        while box.get_last_child() is not None:
            box.remove(box.get_last_child())

        for param in self.fitting_parameters:
            parameter_box = Graphs.FittingParameterBox.new(param)
            parameter_box.set_bounds_visible(
                self.get_settings().get_string("optimization") != "lm",
            )
            for prop in ("initial", "upper_bound", "lower_bound"):
                entry = parameter_box.get_property(prop)
                entry.connect("notify::text", self.on_entry_change)
            box.append(parameter_box)
        return bool(fit)

    def on_entry_change(self, entry, _param) -> None:
        """
        Triggered whenever an entry changes.

        Update the parameters of the curve and perform a new subsequent fit.
        """
        error = False

        def _is_float(value):
            """
            Check if a value can be converted to a float.

            If value cannot be parsed as a float, the CSS class "error" to
            is added to the entry.
            """
            try:
                float(value)
                return True
            except ValueError:
                entry.add_css_class("error")
                return False

        entries = entry.get_ancestor(Graphs.FittingParameterBox)
        # Set the parameters for the row corresponding to the entry that
        # was edited
        for row, params in \
                zip(self.get_fitting_params_box(), self.fitting_parameters):
            if row == entries:
                initial = entries.get_initial().get_text()
                lower_bound = entries.get_lower_bound().get_text()
                upper_bound = entries.get_upper_bound().get_text()
                entries.get_initial().remove_css_class("error")
                entries.get_lower_bound().remove_css_class("error")
                entries.get_upper_bound().remove_css_class("error")

                for bound in [initial, lower_bound, upper_bound]:
                    if not _is_float(bound):
                        self.set_results(error="value")
                        return

                initial = float(initial)
                lower_bound = float(lower_bound)
                upper_bound = float(upper_bound)

                if (initial < lower_bound or initial > upper_bound):
                    entries.get_initial().add_css_class("error")
                    self.set_results(error="bounds")
                    error = True
                if not lower_bound < upper_bound:
                    entries.get_lower_bound().add_css_class("error")
                    entries.get_upper_bound().add_css_class("error")
                    error = True
                    self.set_results(error="bounds")

                params.set_initial(initial)
                params.set_lower_bound(str(lower_bound))
                params.set_upper_bound(str(upper_bound))

        if not error:
            self.fit_curve()

    def set_results(self, error="") -> None:
        """Set the results dialog based on the fit."""
        initial_string = _("Results:") + "\n"
        buffer_string = initial_string
        if error == "value":
            buffer_string += _(
                "Please enter valid fitting \nparameters to start the fit",
            )
        elif error == "equation":
            buffer_string += _(
                "Please enter a valid equation \nto start the fit",
            )
        elif error == "bounds":
            buffer_string += _(
                "Please enter valid fitting bounds \nto start the fit",
            )
        else:
            free_variables = utilities.get_free_variables(
                self.get_equation_string(),
            )
            for index, arg in enumerate(free_variables):
                parameter = utilities.sig_fig_round(self.param[index], 3)
                sigma = utilities.sig_fig_round(self.sigma[index], 3)
                buffer_string += f"{arg}: {parameter}"
                if self.get_settings().get_enum("confidence") != 0:
                    buffer_string += f" (± {sigma})"
                buffer_string += "\n"
            buffer_string += "\n" + _("Sum of R²: {R2}").format(R2=self.r2)
        self.get_text_view().get_buffer().set_text(buffer_string)
        bold_tag = Gtk.TextTag(weight=700)
        self.get_text_view().get_buffer().get_tag_table().add(bold_tag)

        start_iter = self.get_text_view().get_buffer().get_start_iter()
        end_iter = self.get_text_view().get_buffer().get_start_iter()

        # Highlight first word
        while not end_iter.ends_word() and not end_iter.ends_sentence():
            end_iter.forward_char()
        self.get_text_view().get_buffer().apply_tag(
            bold_tag,
            start_iter,
            end_iter,
        )

    def fit_curve(self, *_args) -> bool:
        """
        Fit the data to the equation in the entry.

        Returns a boolean indicating whether the fit was succesfull or not.
        """

        def _get_equation_name(equation_name, values):
            """Obtain the equation name with the fitted parameter values."""
            free_variables = utilities.get_free_variables(
                self.get_equation_string(),
            )
            var_to_val = dict(zip(free_variables, values))
            for var, val in var_to_val.items():
                if var.lower() == "e":
                    pattern = r"(?<!\d)[Ee]|(?!\d)[Ee](?![-+]?\d)"
                else:
                    pattern = (
                        r"((?<=[\d\)])|\b)" + var
                        + r"((?=[\u2070-\u209f\u00b0-\u00be])|\b)"
                    )
                value = utilities.sig_fig_round(val, 3)
                equation_name = \
                    re.sub(pattern, f"({value})", equation_name)
            return equation_name

        function = utilities.string_to_function(self.get_equation_string())
        if function is None:
            return
        try:
            self.param, self.param_cov = curve_fit(
                function,
                self.data_curve.xdata, self.data_curve.ydata,
                p0=self.fitting_parameters.get_p0(),
                bounds=self.fitting_parameters.get_bounds(),
                nan_policy="omit",
                method=self.get_settings().get_string("optimization"),
            )
        except (ValueError, TypeError, _minpack.error, RuntimeError):
            # Cancel fit if not successful
            self.set_results(error="equation")
            return
        xdata = numpy.linspace(
            min(self.data_curve.xdata),
            max(self.data_curve.xdata),
            5000,
        )
        ydata = [function(x, *self.param) for x in xdata]

        name = _get_equation_name(
            str(self.get_custom_equation().get_text()).lower(),
            self.param,
        )
        self.fitted_curve.set_name(f"Y = {name}")
        self.fitted_curve.ydata, self.fitted_curve.xdata = (ydata, xdata)
        self.get_confidence(function)
        self.set_results()
        return True

    def get_confidence(self, function: str) -> None:
        """
        Obtain the confidence interval from the fit.

        Get the confidence interval in terms of the standard deviation
        resulting from the covariance in the plot. The bounds are navively
        calculated by using the extremes on each parameter for both sides
        of the bounds.
        """
        # Get standard deviation
        self.get_canvas().axes[0].relim()  # Reset limits
        self.sigma = numpy.sqrt(numpy.diagonal(self.param_cov))
        self.sigma *= self.get_settings().get_enum("confidence")
        try:
            fitted_y = [
                function(x, *self.param) for x in self.data_curve.xdata
            ]
        except (OverflowError, ZeroDivisionError):
            return
        ss_res = numpy.sum(
            (numpy.asarray(self.data_curve.ydata)
             - numpy.asarray(fitted_y))**2,
        )
        ss_sum = numpy.sum((self.data_curve.ydata - numpy.mean(fitted_y))**2)
        self.r2 = utilities.sig_fig_round(1 - (ss_res / ss_sum), 3)

        # Get confidence band
        upper_bound = function(
            self.fitted_curve.xdata,
            *(self.param + self.sigma),
        )
        lower_bound = function(
            self.fitted_curve.xdata,
            *(self.param - self.sigma),
        )

        # Filter non-finite values from the bounds
        lower_bound = lower_bound[numpy.isfinite(lower_bound)]
        upper_bound = upper_bound[numpy.isfinite(upper_bound)]

        # Cancel if there's no valid values left after filtering non-finite
        if len(upper_bound) == 0 or len(lower_bound) == 0:
            return

        span = max(self.fitted_curve.ydata) - min(self.fitted_curve.ydata)
        middle = \
            (max(self.fitted_curve.ydata) - min(self.fitted_curve.ydata)) / 2

        # Don't try to draw complicated and resource-hogging bounds when
        # far out of range, instead set them to be single-values far away
        if max(upper_bound) > middle + 1e5 * span:
            upper_bound = [middle + 1e5 * span]
        if min(lower_bound) < middle - 1e5 * span:
            lower_bound = [middle - 1e5 * span]

        self.fill.props.data = (
            self.fitted_curve.props.xdata,
            lower_bound,
            upper_bound,
        )

    @staticmethod
    def add_fit(self) -> None:
        """Add fitted data to the items in the main application."""
        application = self.props.application
        style_manager = application.get_figure_style_manager()
        application.get_data().add_items(
            [
                DataItem.new(
                    style_manager.get_selected_style_params(),
                    name=self.fitted_curve.get_name(),
                    xdata=list(self.fitted_curve.xdata),
                    ydata=list(self.fitted_curve.ydata),
                ),
            ],
            style_manager,
        )
        self.close()


class FittingParameterContainer():
    """
    Container class for the Fitting Parameters.

    Each item in the container represents one fitting parameter.
    """

    def __init__(self):
        self._items = {}

    def __iter__(self):
        """Iterate over items."""
        return iter(self._items.values())

    def get(self, arg):
        """Retrieve an item."""
        return self._items[arg]

    def update(self, parameters: list) -> None:
        """
        Remove unused fitting parameters.

        Removes the used parameters `used_list` from the container, this
        ensures that the fitting parameters are kept up-to-date with the
        actual equation.
        """
        names = [item.get_name() for item in self]
        for var in parameters:
            if var not in names:
                self._items[var] = Graphs.FittingParameter(
                    name=var,
                    initial=1,
                    lower_bound="-inf",
                    upper_bound="inf",
                )
        # First create list with items to remove
        # to avoid dict changing size during iteration
        remove_list = []
        for item in self._items.values():
            if item.get_name() not in parameters:
                remove_list.append(item.get_name())

        for item_name in remove_list:
            del self._items[item_name]

        self._items = {key: self._items[key] for key in parameters}

    def get_p0(self) -> list:
        """Get the initial values of the fitting."""
        return [float(item_.get_initial()) for item_ in self]

    def get_bounds(self) -> tuple:
        """Get the bounds of the fitting parameters."""
        lower_bounds = [float(item_.get_lower_bound()) for item_ in self]
        upper_bounds = [float(item_.get_upper_bound()) for item_ in self]
        return (lower_bounds, upper_bounds)



# SPDX-License-Identifier: GPL-3.0-or-later
"""Main actions."""
import logging
import sys
from gettext import gettext as _

from gi.repository import Adw, Gio, Graphs

from graphs import file_io, project, ui, utilities
from graphs.figure_settings import FigureSettingsDialog
from graphs.item import DataItem

import numexpr

import numpy


def on_action_invoked(application: Graphs.Application, name: str) -> None:
    """Handle action invokation."""
    getattr(sys.modules[__name__], name + "_action")(application)


def quit_action(application: Graphs.Application) -> None:
    """Quit the application."""
    application.close_application()


def about_action(application: Graphs.Application) -> None:
    """Display about dialog."""
    Graphs.show_about_dialog(application)


def figure_settings_action(application: Graphs.Application) -> None:
    """Open the figure settings."""
    FigureSettingsDialog(application)


def add_data_action(application: Graphs.Application) -> None:
    """Import data."""
    ui.add_data_dialog(application)


def add_equation_action(application: Graphs.Application) -> None:
    """Add data from an equation."""

    def on_accept(_dialog, name):
        settings = application.get_settings_child("add-equation")
        try:
            x_start = utilities.string_to_float(settings.get_string("x-start"))
            x_stop = utilities.string_to_float(settings.get_string("x-stop"))
            step_size = utilities.string_to_float(
                settings.get_string("step-size"),
            )
            datapoints = int(abs(x_start - x_stop) / step_size) + 1
            xdata = numpy.ndarray.tolist(
                numpy.linspace(x_start, x_stop, datapoints),
            )
            equation = utilities.preprocess(settings.get_string("equation"))
            ydata = numpy.ndarray.tolist(
                numexpr.evaluate(equation + " + x*0", local_dict={"x": xdata}),
            )
            if name == "":
                name = f"Y = {settings.get_string('equation')}"
            style_manager = application.get_figure_style_manager()
            application.get_data().add_items(
                [
                    DataItem.new(
                        style_manager.get_selected_style_params(),
                        xdata,
                        ydata,
                        name=name,
                    ),
                ],
                style_manager,
            )
            return ""
        except ValueError as error:
            return str(error)
        except (NameError, SyntaxError, TypeError, KeyError) as exception:
            message = _("{error} - Unable to add data from equation")
            msg = message.format(error=exception.__class__.__name__)
            logging.exception(msg)
            return msg

    def on_entry_change(entry, _param):
        ui.validate_entry(entry)

    dialog = Graphs.AddEquationDialog.new(application)
    for s in ("x_start", "x_stop", "step_size"):
        entry = dialog.get_property(s)
        entry.connect("notify::text", on_entry_change)
        on_entry_change(entry, None)
    dialog.connect("accept", on_accept)


def select_all_action(application: Graphs.Application) -> None:
    """Select all Items."""
    data = application.get_data()
    for item in data:
        item.set_selected(True)
    data.add_history_state()


def select_none_action(application: Graphs.Application) -> None:
    """Deselect all items."""
    data = application.get_data()
    for item in data:
        item.set_selected(False)
    data.add_history_state()


def undo_action(application: Graphs.Application) -> None:
    """Undo last action."""
    application.get_data().undo()


def redo_action(application: Graphs.Application) -> None:
    """Redo last action."""
    application.get_data().redo()


def optimize_limits_action(application: Graphs.Application) -> None:
    """Optimize figure limits."""
    application.get_data().optimize_limits()


def view_back_action(application: Graphs.Application) -> None:
    """Restore last view."""
    data = application.get_data()
    if data.props.can_view_back:
        data.view_back()


def view_forward_action(application: Graphs.Application) -> None:
    """Restore current view."""
    data = application.get_data()
    if data.props.can_view_forward:
        data.view_forward()


def export_data_action(application: Graphs.Application) -> None:
    """Export Data."""
    ui.export_data_dialog(application)


def export_figure_action(application: Graphs.Application) -> None:
    """Export Figure."""

    def on_accept(_dialog, file):
        window = application.get_window()
        settings = application.get_settings_child("export-figure")
        with file_io.open_wrapped(file, "wb") as wrapper:
            window.get_canvas().figure.savefig(
                wrapper,
                format=settings.get_string("file-format"),
                dpi=settings.get_int("dpi"),
                transparent=settings.get_boolean("transparent"),
            )
        window.add_toast_string_with_file(_("Exported Figure"), file)

    dialog = Graphs.ExportFigureDialog.new(application)
    dialog.connect("accept", on_accept)


def new_project_action(application: Graphs.Application) -> None:
    """Clear the current project and reset Graphs to the initial state."""
    data = application.get_data()
    if data.get_unsaved():

        def on_response(_dialog, response):
            application.save_handler = data.connect(
                "saved",
                application.on_project_saved,
                "reset_project",
            )
            if response == "discard_close":
                data.reset()
            if response == "save_close":
                project.save_project(application)

        dialog = Graphs.tools_build_dialog("save_changes")
        dialog.connect("response", on_response)
        dialog.present(application.get_window())
        return
    data.reset()


def save_project_action(application: Graphs.Application) -> None:
    """Save the current project."""
    project.save_project(application)


def save_project_as_action(application: Graphs.Application) -> None:
    """Save the current project and ask for save location."""
    project.save_project(application, require_dialog=True)


def smoothen_settings_action(application: Graphs.Application) -> None:
    """Open settings for the smoothen action."""
    Graphs.SmoothenDialog.new(application)


def zoom_in_action(application: Graphs.Application) -> None:
    """Zoom into the figure."""
    canvas = application.get_window().get_canvas()
    canvas.zoom(1.15, respect_mouse=False)


def zoom_out_action(application: Graphs.Application) -> None:
    """Zoom out of the figure."""
    canvas = application.get_window().get_canvas()
    canvas.zoom(1 / 1.15, respect_mouse=False)


def open_project_action(application: Graphs.Application) -> None:
    """Open a project."""
    project.open_project(application)


def delete_selected_action(application: Graphs.Application) -> None:
    """Delete selected items."""
    items = [item for item in application.get_data() if item.get_selected()]
    names = ", ".join(item.get_name() for item in items)
    application.get_data().delete_items(items)
    toast = Adw.Toast.new(_("Deleted {name}").format(name=names))
    toast.set_button_label(_("Undo"))
    toast.set_action_name("app.undo")
    application.get_window().add_toast(toast)


def help_action(application: Graphs.Application) -> None:
    """Open Help page for Graphs."""
    Gio.AppInfo.launch_default_for_uri(
        "help:graphs",
        application.get_window().get_display().get_app_launch_context(),
    )



# SPDX-License-Identifier: GPL-3.0-or-later
"""
Wrapper classes for mpl artists.

Provides GObject based wrappers for mpl artists.
"""
from gi.repository import GObject, Graphs

from graphs import misc

from matplotlib import artist, pyplot
from matplotlib.figure import Figure


def new_for_item(canvas: Graphs.Canvas, item: Graphs.Item):
    """
    Create a new artist for an item.

    Creates bindings between item and artist properties so changes are handled
    automatically.
    """
    match item.__gtype_name__:
        case "GraphsDataItem":
            cls = DataItemArtistWrapper
        case "GraphsTextItem":
            cls = TextItemArtistWrapper
        case "GraphsFillItem":
            cls = FillItemArtistWrapper
        case _:
            pass
    artist_wrapper = cls(
        canvas.axes[item.get_yposition() * 2 + item.get_xposition()],
        item,
    )
    for prop in dir(artist_wrapper.props):
        if not (prop == "label" and artist_wrapper.legend):
            item.bind_property(prop, artist_wrapper, prop, 0)
    artist_wrapper.connect("notify", lambda _x, _y: canvas.update_legend())
    return artist_wrapper


class ItemArtistWrapper(GObject.Object):
    """Wrapper for base Item."""

    __gtype_name__ = "GraphsItemArtistWrapper"
    legend = False

    def get_artist(self) -> artist:
        """Get underlying mpl artist."""
        return self._artist

    @GObject.Property(type=str, default="")
    def name(self) -> str:
        """Get name/label property."""
        return self._artist.get_label()

    @name.setter
    def name(self, name: str) -> None:
        """Set name/label property."""
        self._artist.set_label(Graphs.tools_shorten_label(name, 40))

    @GObject.Property(type=str, default="000000")
    def color(self) -> str:
        """Get color property."""
        return self._artist.get_color()

    @color.setter
    def color(self, color: str) -> None:
        """Set color property."""
        self._artist.set_color(color)

    @GObject.Property(type=float, default=1)
    def alpha(self) -> float:
        """Get alpha property."""
        return self._artist.get_alpha()

    @alpha.setter
    def alpha(self, alpha: float) -> None:
        """Set alpha property."""
        self._artist.set_alpha(alpha)


class DataItemArtistWrapper(ItemArtistWrapper):
    """Wrapper for DataItem."""

    __gtype_name__ = "GraphsDataItemArtistWrapper"
    selected = GObject.Property(type=bool, default=True)
    linewidth = GObject.Property(type=float, default=3)
    markersize = GObject.Property(type=float, default=7)
    legend = True

    @GObject.Property
    def xdata(self) -> list:
        """Get xdata property."""
        return self._artist.get_xdata()

    @xdata.setter
    def xdata(self, xdata: list) -> None:
        """Set xdata property."""
        self._artist.set_xdata(xdata)

    @GObject.Property
    def ydata(self) -> list:
        """Get ydata property."""
        return self._artist.get_ydata()

    @ydata.setter
    def ydata(self, ydata: list) -> None:
        """Set ydata property."""
        self._artist.set_ydata(ydata)

    @GObject.Property(type=int, default=1)
    def linestyle(self) -> int:
        """Get linestyle property."""
        return misc.LINESTYLES.index(self._artist.get_linestyle())

    @linestyle.setter
    def linestyle(self, linestyle: int) -> None:
        """Set linestyle property."""
        self._artist.set_linestyle(misc.LINESTYLES[linestyle])

    @GObject.Property(type=int, default=1)
    def markerstyle(self) -> int:
        """Get markerstyle property."""
        return misc.MARKERSTYLES.index(self._artist.get_marker())

    @markerstyle.setter
    def markerstyle(self, markerstyle: int) -> None:
        """Set markerstyle property."""
        self._artist.set_marker(misc.MARKERSTYLES[markerstyle])

    def _set_properties(self, _x, _y) -> None:
        linewidth, markersize = self.props.linewidth, self.props.markersize
        if not self.props.selected:
            linewidth *= 0.35
            markersize *= 0.35
        self._artist.set_linewidth(linewidth)
        self._artist.set_markersize(markersize)

    def __init__(self, axis: pyplot.axis, item: Graphs.Item):
        super().__init__()
        self._artist = axis.plot(
            item.props.xdata,
            item.props.ydata,
            label=Graphs.tools_shorten_label(item.get_name(), 40),
            color=item.get_color(),
            alpha=item.get_alpha(),
            linestyle=misc.LINESTYLES[item.props.linestyle],
            marker=misc.MARKERSTYLES[item.props.markerstyle],
        )[0]
        for prop in ("selected", "linewidth", "markersize"):
            self.set_property(prop, item.get_property(prop))
            self.connect(f"notify::{prop}", self._set_properties)
        self._set_properties(None, None)


class TextItemArtistWrapper(ItemArtistWrapper):
    """Wrapper for TextItem."""

    __gtype_name__ = "GraphsTextItemArtistWrapper"

    @GObject.Property(type=float, default=12)
    def size(self) -> float:
        """Get size property."""
        return self._artist.get_fontsize()

    @size.setter
    def size(self, size: float) -> None:
        """Set size property."""
        self._artist.set_fontsize(size)

    @GObject.Property(type=int, default=0, minimum=0, maximum=360)
    def rotation(self) -> int:
        """Get rotation property."""
        return self._artist.get_rotation()

    @rotation.setter
    def rotation(self, rotation: int) -> None:
        """Set rotation property."""
        self._artist.set_rotation(rotation)

    @GObject.Property(type=str, default="")
    def text(self) -> str:
        """Get text property."""
        return self._artist.get_text()

    @text.setter
    def text(self, text: str) -> None:
        """Set text property."""
        self._artist.set_text(text)

    @GObject.Property(type=float, default=0)
    def xanchor(self) -> float:
        """Get xanchor property."""
        return self._artist.get_position()[0]

    @xanchor.setter
    def xanchor(self, xanchor: float) -> None:
        """Set xanchor property."""
        self._artist.set_position((xanchor, self.props.yanchor))

    @GObject.Property(type=float, default=0)
    def yanchor(self) -> float:
        """Get yanchor property."""
        return self._artist.get_position()[1]

    @yanchor.setter
    def yanchor(self, yanchor: float) -> None:
        """Set yanchor property."""
        self._artist.set_position((self.props.xanchor, yanchor))

    def __init__(self, axis: pyplot.axis, item: Graphs.Item):
        super().__init__()
        self._artist = axis.text(
            item.props.xanchor,
            item.props.yanchor,
            item.props.text,
            label=Graphs.tools_shorten_label(item.get_name(), 40),
            color=item.get_color(),
            alpha=item.get_alpha(),
            clip_on=True,
            fontsize=item.props.size,
            rotation=item.props.rotation,
        )


class FillItemArtistWrapper(ItemArtistWrapper):
    """Wrapper for FillItem."""

    __gtype_name__ = "GraphsFillItemArtistWrapper"

    @GObject.Property(type=object, flags=2)
    def data(self) -> None:
        """Write-only property, ignored."""

    @data.setter
    def data(self, data) -> None:
        dummy = Figure().add_subplot().fill_between(*data)
        self._artist.set_paths([dummy.get_paths()[0].vertices])

    def __init__(self, axis: pyplot.axis, item: Graphs.Item):
        super().__init__()
        self._artist = axis.fill_between(
            *item.props.data,
            label=Graphs.tools_shorten_label(item.get_name(), 40),
            color=item.get_color(),
            alpha=item.get_alpha(),
        )
